const coral = @import("../coral.zig");

const std = @import("std");

const tokens = @import("./tokens.zig");

scope: *coral.shaders.Scope,
inputs: std.BoundedArray(*const coral.shaders.Input, 15) = .{},
outputs: std.BoundedArray(*const coral.shaders.Output, 15) = .{},
uniforms: std.BoundedArray(*const coral.shaders.Uniform, 15) = .{},
textures: std.BoundedArray(*const coral.shaders.Texture, 15) = .{},
functions: std.BoundedArray(*const coral.shaders.Function, max_functions) = .{},

pub const ParseResult = union(enum) {
    ok,
    failure: [:0]const u8,
};

const Self = @This();

pub fn clear(self: *Self) void {
    self.scope.clear();
    self.inputs.clear();
    self.outputs.clear();
    self.uniforms.clear();
    self.textures.clear();
    self.functions.clear();
}

pub fn defineFunction(self: *Self, arena: *std.heap.ArenaAllocator, function: coral.shaders.Function) coral.shaders.DefinitionError!void {
    self.functions.append(try self.scope.define(arena, function)) catch |append_error| {
        return switch (append_error) {
            error.Overflow => error.TooManySymbols,
        };
    };
}

pub fn defineInput(self: *Self, arena: *std.heap.ArenaAllocator, input: coral.shaders.Input) coral.shaders.DefinitionError!void {
    self.inputs.append(try self.scope.define(arena, input)) catch |append_error| {
        return switch (append_error) {
            error.Overflow => error.TooManySymbols,
        };
    };
}

pub fn defineOutput(self: *Self, arena: *std.heap.ArenaAllocator, output: coral.shaders.Output) coral.shaders.DefinitionError!void {
    self.outputs.append(try self.scope.define(arena, output)) catch |append_error| {
        return switch (append_error) {
            error.Overflow => error.TooManySymbols,
        };
    };
}

pub fn defineTexture(self: *Self, arena: *std.heap.ArenaAllocator, texture: coral.shaders.Texture) coral.shaders.DefinitionError!void {
    self.textures.append(try self.scope.define(arena, texture)) catch |append_error| {
        return switch (append_error) {
            error.Overflow => error.TooManySymbols,
        };
    };
}

pub fn defineUniform(self: *Self, arena: *std.heap.ArenaAllocator, uniform: coral.shaders.Uniform) coral.shaders.DefinitionError!void {
    self.uniforms.append(try self.scope.define(arena, uniform)) catch |append_error| {
        return switch (append_error) {
            error.Overflow => error.TooManySymbols,
        };
    };
}

pub fn hasFunction(self: Self, identifier: []const u8) ?*const coral.shaders.Function {
    return self.scope.hasLocal(coral.shaders.Function, identifier);
}

pub fn hasInput(self: Self, identifier: []const u8) ?*const coral.shaders.Input {
    return self.scope.hasLocal(coral.shaders.Input, identifier);
}

pub fn hasOutput(self: Self, identifier: []const u8) ?*const coral.shaders.Output {
    return self.scope.hasLocal(coral.shaders.Output, identifier);
}

pub fn hasTexture(self: Self, identifier: []const u8) ?*const coral.shaders.Texture {
    return self.scope.hasLocal(coral.shaders.Texture, identifier);
}

pub fn hasUniform(self: Self, identifier: []const u8) ?*const coral.shaders.Uniform {
    return self.scope.hasLocal(coral.shaders.Uniform, identifier);
}

pub fn init(arena: *std.heap.ArenaAllocator) std.mem.Allocator.Error!Self {
    const scope = try arena.allocator().create(coral.shaders.Scope);

    scope.* = .{};

    return .{ .scope = scope };
}

const max_functions = 255;

pub fn parse(self: *Self, arena: *std.heap.ArenaAllocator, source_text: []const u8) std.mem.Allocator.Error!ParseResult {
    errdefer {
        self.clear();
    }

    var source = tokens.Stream.init(source_text);

    self.parseDocument(arena, &source) catch |parse_error| {
        const arena_allocator = arena.allocator();

        return .{
            .failure = try switch (parse_error) {
                error.OutOfMemory => error.OutOfMemory,

                error.TooManyConstants => coral.bytes.allocFormatted(arena_allocator, "{location}: number of literals in the current scope exceeded", .{
                    .location = source.location,
                }),

                error.ImmutableStorage => coral.bytes.allocFormatted(arena_allocator, "{location}: attempt to modify an immutable value", .{
                    .location = source.location,
                }),

                error.TooManySymbols => coral.bytes.allocFormatted(arena_allocator, "{location}: number of definitions in the current scope exceeded", .{
                    .location = source.location,
                }),

                error.IncompatibleArguments => coral.bytes.allocFormatted(arena_allocator, "{location}: incompatible set of arguments", .{
                    .location = source.location,
                }),

                error.IncompatibleTypes => coral.bytes.allocFormatted(arena_allocator, "{location}: incompatible types", .{
                    .location = source.location,
                }),

                error.DuplicateIdentifier => coral.bytes.allocFormatted(arena_allocator, "{location}: {kind} {token} already defined", .{
                    .location = source.location,
                    .token = source.current,
                    .kind = std.meta.activeTag(source.current),
                }),

                error.UnexpectedToken => coral.bytes.allocFormatted(arena_allocator, "{location}: unexpected {token}", .{
                    .location = source.location,
                    .token = source.current,
                }),

                error.UndefinedIdentifier => coral.bytes.allocFormatted(arena_allocator, "{location}: undefined identifier {token}", .{
                    .location = source.location,
                    .token = source.current,
                }),

                error.MissingReturn => coral.bytes.allocFormatted(arena_allocator, "{location}: value-returning function does not return at end", .{
                    .location = source.location,
                }),
            },
        };
    };

    return .ok;
}

fn parseDocument(self: *Self, arena: *std.heap.ArenaAllocator, source: *tokens.Stream) coral.shaders.ParsingError!void {
    while (source.skip(.newline) != .end) {
        try switch (try source.current.expectAny(&.{.keyword_function})) {
            .keyword_function => self.parseFunction(arena, source),
        };
    }
}

fn parseFunction(self: *Self, arena: *std.heap.ArenaAllocator, source: *tokens.Stream) coral.shaders.ParsingError!void {
    const identifier = try arena.allocator().dupeZ(u8, try source.skip(.newline).expectIdentifier());

    try source.skip(.newline).expect(.symbol_paren_left);

    const inner_scope = try self.scope.createScope(arena);

    const parameters = switch (try source.skip(.newline).expectAny(&.{ .symbol_paren_right, .identifier })) {
        .identifier => try inner_scope.parseParameter(arena, source),
        .symbol_paren_right => null,
    };

    const parameter_count = inner_scope.defined;

    try source.skip(.newline).expect(.symbol_arrow);

    var peeking = source.*;
    const has_return_type = if (peeking.skip(.newline) == .symbol_brace_left) null else try self.scope.parseType(source);
    const block = try inner_scope.parseBlock(arena, source);

    if (has_return_type) |return_type| {
        const last_statement = block.hasLastStatement() orelse {
            return error.MissingReturn;
        };

        if (last_statement.* != .return_expression) {
            return error.MissingReturn;
        }

        if (try last_statement.return_expression.inferType() != return_type) {
            return error.IncompatibleTypes;
        }
    }

    try self.defineFunction(arena, .{
        .has_return_type = has_return_type,
        .parameters = parameters,
        .parameter_count = parameter_count,
        .identifier = identifier,
        .block = block,
        .signature = try self.scope.intern(arena, try signature(arena, parameters, has_return_type)),
    });
}

fn signature(arena: *std.heap.ArenaAllocator, parameters: ?*const coral.shaders.Parameter, has_return_type: ?*const coral.shaders.Type) std.mem.Allocator.Error![:0]const u8 {
    var buffer_size = 2 + (if (has_return_type) |return_type| return_type.identifier.len else 0);

    if (parameters) |first_parameter| {
        buffer_size += first_parameter.type.identifier.len;

        var has_parameter = first_parameter.has_next;

        while (has_parameter) |parameter| : (has_parameter = parameter.has_next) {
            buffer_size += 1 + parameter.type.identifier.len;
        }
    }

    var buffer = coral.bytes.span(try arena.allocator().allocSentinel(u8, buffer_size, 0));

    std.debug.assert(buffer.put('('));

    if (parameters) |first_parameter| {
        std.debug.assert(buffer.write(first_parameter.type.identifier) == first_parameter.type.identifier.len);

        var has_parameter = first_parameter.has_next;

        while (has_parameter) |parameter| : (has_parameter = parameter.has_next) {
            std.debug.assert(buffer.put(','));
            std.debug.assert(buffer.write(parameter.type.identifier) == parameter.type.identifier.len);
        }
    }

    std.debug.assert(buffer.put(')'));

    if (has_return_type) |return_type| {
        std.debug.assert(buffer.write(return_type.identifier) == return_type.identifier.len);
    }

    return @ptrCast(buffer.bytes);
}
