const coral = @import("../coral.zig");

const std = @import("std");

const tokens = @import("./tokens.zig");

interned: coral.tree.Binary([:0]const u8, void, coral.tree.sliceTraits([:0]const u8)) = .empty,
identifiers: [max_definitions][:0]const u8 = undefined,
symbols: [max_definitions]coral.Box = undefined,
defined: usize = 0,
has_enclosing: ?*Self = null,

const Self = @This();

pub fn clear(self: *Self) void {
    self.defined = 0;

    if (self.has_enclosing) |enclosing| {
        enclosing.clear();
    }
}

fn create(arena: *std.heap.ArenaAllocator, node: anytype) std.mem.Allocator.Error!*@TypeOf(node) {
    const allocation = try arena.allocator().create(@TypeOf(node));

    allocation.* = node;

    return allocation;
}

pub fn createScope(self: *Self, arena: *std.heap.ArenaAllocator) std.mem.Allocator.Error!*Self {
    return create(arena, Self{
        .has_enclosing = self,
    });
}

pub fn define(self: *Self, arena: *std.heap.ArenaAllocator, symbol: anytype) coral.shaders.DefinitionError!*@TypeOf(symbol) {
    const Symbol = @TypeOf(symbol);
    const identifier_field = "identifier";
    const Identifier = [:0]const u8;

    const identifier = switch (@hasField(Symbol, identifier_field)) {
        true => switch (@TypeOf(symbol.identifier)) {
            Identifier => symbol.identifier,

            else => @compileError(std.fmt.comptimePrint("field {s}.{s} must be of type `{s}`", .{
                @typeName(Symbol),
                identifier_field,
                @typeName(Identifier),
            })),
        },

        false => @compileError(std.fmt.comptimePrint("Type `{s}` must contain a field identifierd `{s}`", .{
            @typeName(Symbol),
            identifier_field,
        })),
    };

    if (self.definitionExists(identifier)) {
        return error.DuplicateIdentifier;
    }

    if (self.defined >= max_definitions) {
        return error.TooManySymbols;
    }

    const stored_symbol = &self.symbols[self.defined];

    self.identifiers[self.defined] = identifier;
    stored_symbol.* = try .initWithAllocator(arena.allocator(), symbol);
    self.defined += 1;

    return stored_symbol.has(Symbol).?;
}

fn definitionExists(self: Self, identifier: []const u8) bool {
    for (self.identifiers[0..self.defined]) |existing_identifier| {
        if (std.mem.eql(u8, existing_identifier, identifier)) {
            return true;
        }
    }

    if (self.has_enclosing) |enclosing| {
        return enclosing.definitionExists(identifier);
    }

    return false;
}

pub fn hasGlobal(self: *const Self, comptime Symbol: type, identifier: []const u8) ?*const Symbol {
    if (self.hasLocal(Symbol, identifier)) |local| {
        return local;
    }

    if (self.has_enclosing) |enclosing| {
        return enclosing.hasGlobal(Symbol, identifier);
    }

    return null;
}

pub fn hasLocal(self: *const Self, comptime Symbol: type, identifier: []const u8) ?*const Symbol {
    for (0..self.defined) |i| {
        if (std.mem.eql(u8, self.identifiers[i], identifier)) {
            if (self.symbols[i].has(Symbol)) |symbol| {
                return symbol;
            }
        }
    }

    return null;
}

pub fn intern(self: *Self, arena: *std.heap.ArenaAllocator, value: [:0]const u8) std.mem.Allocator.Error![:0]const u8 {
    var scope: ?*Self = self;

    while (scope) |current| : (scope = current.has_enclosing) {
        if (current.interned.getKey(value)) |existing| {
            return existing;
        }
    }

    std.debug.assert(try self.interned.insert(arena.allocator(), value, {}) != null);

    return value;
}

const max_definitions = 256;

fn parseArgument(self: *Self, arena: *std.heap.ArenaAllocator, source: *tokens.Stream) coral.shaders.ParsingError!*const coral.shaders.Argument {
    const expression = try self.parseExpression(arena, source);

    return switch (try source.current.expectAny(&.{ .symbol_paren_right, .symbol_comma })) {
        .symbol_paren_right => try create(arena, coral.shaders.Argument{
            .expression = expression,
        }),

        .symbol_comma => {
            var peeking = source.*;

            if (peeking.skip(.newline) == .symbol_paren_right) {
                source.* = peeking;

                return try create(arena, coral.shaders.Argument{ .expression = expression });
            }

            return try create(arena, coral.shaders.Argument{
                .expression = expression,
                .has_next = try self.parseArgument(arena, source),
            });
        },
    };
}

pub fn parseBlock(self: *Self, arena: *std.heap.ArenaAllocator, source: *tokens.Stream) coral.shaders.ParsingError!*const coral.shaders.Block {
    try source.skip(.newline).expect(.symbol_brace_left);

    return create(arena, coral.shaders.Block{
        .scope = self,
        .depth = source.depth,

        .statements = switch (source.skip(.newline)) {
            .symbol_brace_right => null,
            else => try self.parseBlockStatement(arena, source),
        },
    });
}

fn parseBlockStatement(self: *Self, arena: *std.heap.ArenaAllocator, source: *tokens.Stream) coral.shaders.ParsingError!*const coral.shaders.Statement {
    switch (try source.current.expectAny(&.{ .keyword_let, .keyword_var, .keyword_return, .identifier })) {
        .identifier => {
            const identifier = try source.current.expectIdentifier();

            if (self.hasLocal(coral.shaders.Local, identifier)) |local| {
                if (local.is_constant) {
                    return error.ImmutableStorage;
                }

                try source.skip(.newline).expect(.symbol_equals);

                const expression = try self.parseExpression(arena, source);

                if (try expression.inferType() != local.type) {
                    return error.IncompatibleTypes;
                }

                return try create(arena, coral.shaders.Statement{
                    .mutate_local = .{
                        .local = local,
                        .expression = expression,

                        .has_next = switch (source.current) {
                            .symbol_brace_right => null,
                            else => try self.parseBlockStatement(arena, source),
                        },
                    },
                });
            }

            if (self.hasGlobal(coral.shaders.Output, identifier)) |output| {
                try source.skip(.newline).expect(.symbol_equals);

                const expression = try self.parseExpression(arena, source);

                if (try expression.inferType() != output.type) {
                    return error.IncompatibleTypes;
                }

                return try create(arena, coral.shaders.Statement{
                    .mutate_output = .{
                        .output = output,
                        .expression = expression,

                        .has_next = switch (source.current) {
                            .symbol_brace_right => null,
                            else => try self.parseBlockStatement(arena, source),
                        },
                    },
                });
            }

            return error.ImmutableStorage;
        },

        .keyword_return => {
            return try create(arena, coral.shaders.Statement{
                .return_expression = try self.parseExpression(arena, source),
            });
        },

        .keyword_var, .keyword_let => {
            const local = try self.define(arena, coral.shaders.Local{
                .is_constant = source.current != .keyword_var,
                .identifier = try arena.allocator().dupeZ(u8, try source.next().expectIdentifier()),
                .expression = &.{ .constant = try self.intern(arena, "0") },
                .type = .int,
            });

            try source.skip(.newline).expect(.symbol_equals);

            local.expression = try self.parseExpression(arena, source);
            local.type = try local.expression.inferType();

            return try create(arena, coral.shaders.Statement{
                .declare_local = .{
                    .local = local,

                    .has_next = try switch (source.current) {
                        .symbol_brace_right => null,
                        else => self.parseBlockStatement(arena, source),
                    },
                },
            });
        },
    }
}

fn parseExpression(self: *Self, arena: *std.heap.ArenaAllocator, source: *tokens.Stream) coral.shaders.ParsingError!*const coral.shaders.Expression {
    const expression = try self.parseAdditiveExpression(arena, source);

    if (source.current == .symbol_equals) {
        return switch (expression.*) {
            .get_local => |local| switch (local.is_constant) {
                false => try create(arena, coral.shaders.Expression{
                    .mutate_local = .{
                        .local = local,
                        .expression = try self.parseExpression(arena, source),
                    },
                }),

                true => error.ImmutableStorage,
            },

            else => error.UnexpectedToken,
        };
    }

    return expression;
}

fn parseAdditiveExpression(self: *Self, arena: *std.heap.ArenaAllocator, source: *tokens.Stream) coral.shaders.ParsingError!*const coral.shaders.Expression {
    const lhs_expression = try self.parseEqualityExpression(arena, source);

    if (source.current == .symbol_plus) {
        const rhs_expression = try self.parseEqualityExpression(arena, source);
        const lhs_type = try lhs_expression.inferType();

        if (lhs_type != try rhs_expression.inferType()) {
            return error.IncompatibleTypes;
        }

        return create(arena, coral.shaders.Expression{
            .add = .{
                .rhs_expression = rhs_expression,
                .lhs_expression = lhs_expression,
                .type = lhs_type,
            },
        });
    }

    if (source.current == .symbol_minus) {
        const rhs_expression = try self.parseEqualityExpression(arena, source);
        const lhs_type = try lhs_expression.inferType();

        if (lhs_type != try rhs_expression.inferType()) {
            return error.IncompatibleTypes;
        }

        return create(arena, coral.shaders.Expression{
            .subtract = .{
                .rhs_expression = rhs_expression,
                .lhs_expression = lhs_expression,
                .type = lhs_type,
            },
        });
    }

    return lhs_expression;
}

fn parseComparativeExpression(self: *Self, arena: *std.heap.ArenaAllocator, source: *tokens.Stream) coral.shaders.ParsingError!*const coral.shaders.Expression {
    const lhs_expression = try self.parseTermExpression(arena, source);

    if (source.current == .symbol_greater_than) {
        const rhs_expression = try self.parseTermExpression(arena, source);
        const lhs_type = try lhs_expression.inferType();

        if (lhs_type != try rhs_expression.inferType()) {
            return error.IncompatibleTypes;
        }

        return create(arena, coral.shaders.Expression{
            .greater_than = .{
                .rhs_expression = rhs_expression,
                .lhs_expression = lhs_expression,
                .type = lhs_type,
            },
        });
    }

    if (source.current == .symbol_greater_equals) {
        const rhs_expression = try self.parseTermExpression(arena, source);
        const lhs_type = try lhs_expression.inferType();

        if (lhs_type != try rhs_expression.inferType()) {
            return error.IncompatibleTypes;
        }

        return create(arena, coral.shaders.Expression{
            .greater_equal = .{
                .rhs_expression = rhs_expression,
                .lhs_expression = lhs_expression,
                .type = lhs_type,
            },
        });
    }

    if (source.current == .symbol_lesser_than) {
        const rhs_expression = try self.parseTermExpression(arena, source);
        const lhs_type = try lhs_expression.inferType();

        if (lhs_type != try rhs_expression.inferType()) {
            return error.IncompatibleTypes;
        }

        return create(arena, coral.shaders.Expression{
            .divide = .{
                .rhs_expression = rhs_expression,
                .lhs_expression = lhs_expression,
                .type = lhs_type,
            },
        });
    }

    if (source.current == .symbol_lesser_equals) {
        const rhs_expression = try self.parseTermExpression(arena, source);
        const lhs_type = try lhs_expression.inferType();

        if (lhs_type != try rhs_expression.inferType()) {
            return error.IncompatibleTypes;
        }

        return create(arena, coral.shaders.Expression{
            .divide = .{
                .rhs_expression = rhs_expression,
                .lhs_expression = lhs_expression,
                .type = lhs_type,
            },
        });
    }

    return lhs_expression;
}

fn parseEqualityExpression(self: *Self, arena: *std.heap.ArenaAllocator, source: *tokens.Stream) coral.shaders.ParsingError!*const coral.shaders.Expression {
    const lhs_expression = try self.parseComparativeExpression(arena, source);

    if (source.current == .symbol_double_equals) {
        const rhs_expression = try self.parseComparativeExpression(arena, source);
        const lhs_type = try lhs_expression.inferType();

        if (lhs_type != try rhs_expression.inferType()) {
            return error.IncompatibleTypes;
        }

        return create(arena, coral.shaders.Expression{
            .equal = .{
                .rhs_expression = rhs_expression,
                .lhs_expression = lhs_expression,
                .type = lhs_type,
            },
        });
    }

    return lhs_expression;
}

fn parseFactorExpression(self: *Self, arena: *std.heap.ArenaAllocator, source: *tokens.Stream) coral.shaders.ParsingError!*const coral.shaders.Expression {
    var expression = try self.parseOperandExpression(arena, source);

    while (true) {
        switch (source.skip(.newline)) {
            .symbol_period => {
                const object_type = try expression.inferType();

                expression = try create(arena, coral.shaders.Expression{
                    .get_object = .{
                        .object_expression = expression,

                        .field = object_type.hasField(try source.skip(.newline).expectIdentifier()) orelse {
                            return error.UndefinedIdentifier;
                        },
                    },
                });
            },

            else => {
                return expression;
            },
        }
    }
}

fn parseOperandExpression(self: *Self, arena: *std.heap.ArenaAllocator, source: *tokens.Stream) coral.shaders.ParsingError!*const coral.shaders.Expression {
    switch (source.skip(.newline)) {
        .symbol_minus => {
            return try create(arena, coral.shaders.Expression{
                .negate_expression = try self.parseExpression(arena, source),
            });
        },

        .symbol_paren_left => {
            return try create(arena, coral.shaders.Expression{
                .group_expression = try self.parseExpression(arena, source),
            });
        },

        .scalar => {
            return try create(arena, coral.shaders.Expression{
                .constant = try arena.allocator().dupeZ(u8, try source.current.expectScalar()),
            });
        },

        .identifier => |identifier| {
            if (self.hasGlobal(coral.shaders.Function, identifier)) |function| {
                try source.skip(.newline).expect(.symbol_paren_left);

                var peeking = source.*;

                switch (peeking.skip(.newline)) {
                    .symbol_paren_left => {
                        source.* = peeking;

                        return try create(arena, coral.shaders.Expression{
                            .invoke = .{
                                .function = function,
                            },
                        });
                    },

                    else => {
                        const arguments = try self.parseArgument(arena, source);
                        var argument_count: usize = 1;

                        {
                            var subsequent_arguments = arguments.has_next;

                            while (subsequent_arguments) |argument| : (subsequent_arguments = argument.has_next) {
                                argument_count += 1;
                            }
                        }

                        return try create(arena, coral.shaders.Expression{
                            .invoke = .{
                                .function = function,
                                .argument_count = argument_count,
                                .arguments = arguments,
                            },
                        });
                    },
                }
            }

            if (self.hasGlobal(coral.shaders.Uniform, identifier)) |uniform| {
                try source.skip(.newline).expect(.symbol_period);

                const field_name = try source.skip(.newline).expectIdentifier();

                const first_field = uniform.has_field orelse {
                    return error.UndefinedIdentifier;
                };

                return try create(arena, coral.shaders.Expression{
                    .get_uniform = .{
                        .uniform = uniform,

                        .field = first_field.has(field_name) orelse {
                            return error.UndefinedIdentifier;
                        },
                    },
                });
            }

            if (self.hasGlobal(coral.shaders.Texture, identifier)) |texture| {
                return try create(arena, coral.shaders.Expression{ .get_texture = texture });
            }

            if (self.hasGlobal(coral.shaders.Output, identifier)) |output| {
                return try create(arena, coral.shaders.Expression{ .get_output = output });
            }

            if (self.hasGlobal(coral.shaders.Input, identifier)) |input| {
                return try create(arena, coral.shaders.Expression{ .get_input = input });
            }

            if (self.hasLocal(coral.shaders.Local, identifier)) |local| {
                return try create(arena, coral.shaders.Expression{ .get_local = local });
            }

            if (self.hasLocal(coral.shaders.Parameter, identifier)) |parameter| {
                return try create(arena, coral.shaders.Expression{ .get_parameter = parameter });
            }

            return error.UndefinedIdentifier;
        },

        .keyword_pow => {
            try source.skip(.newline).expect(.symbol_paren_left);

            return try create(arena, coral.shaders.Expression{
                .pow = .{
                    .allowed_parameter_types = &.{ .float, .float2, .float3, .float4 },
                    .expected_parameter_count = 2,
                    .first_argument = try self.parseArgument(arena, source),
                },
            });
        },

        .keyword_abs => {
            try source.skip(.newline).expect(.symbol_paren_left);

            return try create(arena, coral.shaders.Expression{
                .abs = .{
                    .allowed_parameter_types = &.{ .float, .float2, .float3, .float4 },
                    .expected_parameter_count = 1,
                    .first_argument = try self.parseArgument(arena, source),
                },
            });
        },

        .keyword_sin => {
            try source.skip(.newline).expect(.symbol_paren_left);

            return try create(arena, coral.shaders.Expression{
                .sin = .{
                    .allowed_parameter_types = &.{ .float, .float2, .float3, .float4 },
                    .expected_parameter_count = 1,
                    .first_argument = try self.parseArgument(arena, source),
                },
            });
        },

        .keyword_int => {
            try source.skip(.newline).expect(.symbol_paren_left);

            return try create(arena, coral.shaders.Expression{
                .convert = .{
                    .target_type = .int,
                    .parameter_types = &.{.float},
                    .first_argument = try self.parseArgument(arena, source),
                },
            });
        },

        .keyword_float => {
            try source.skip(.newline).expect(.symbol_paren_left);

            return try create(arena, coral.shaders.Expression{
                .convert = .{
                    .target_type = .float,
                    .parameter_types = &.{.int},
                    .first_argument = try self.parseArgument(arena, source),
                },
            });
        },

        .keyword_float2 => {
            try source.skip(.newline).expect(.symbol_paren_left);

            return try create(arena, coral.shaders.Expression{
                .convert = .{
                    .target_type = .float2,
                    .parameter_types = &.{ .float, .float },
                    .first_argument = try self.parseArgument(arena, source),
                },
            });
        },

        .keyword_float3 => {
            try source.skip(.newline).expect(.symbol_paren_left);

            return try create(arena, coral.shaders.Expression{
                .convert = .{
                    .target_type = .float3,
                    .parameter_types = &.{ .float, .float, .float },
                    .first_argument = try self.parseArgument(arena, source),
                },
            });
        },

        .keyword_float4 => {
            try source.skip(.newline).expect(.symbol_paren_left);

            return try create(arena, coral.shaders.Expression{
                .convert = .{
                    .target_type = .float4,
                    .parameter_types = &.{ .float, .float, .float, .float },
                    .first_argument = try self.parseArgument(arena, source),
                },
            });
        },

        .keyword_float4x4 => {
            try source.skip(.newline).expect(.symbol_paren_left);

            return try create(arena, coral.shaders.Expression{
                .convert = .{
                    .target_type = .float4x4,
                    .parameter_types = &.{ .float4, .float4, .float4, .float4 },
                    .first_argument = try self.parseArgument(arena, source),
                },
            });
        },

        .keyword_sample => {
            try source.skip(.newline).expect(.symbol_paren_left);

            return try create(arena, coral.shaders.Expression{
                .sample = .{
                    .signatures = &.{
                        .{
                            .parameter_types = &.{ .texture2, .float2 },
                            .return_type = .float4,
                        },
                    },

                    .has_argument = try self.parseArgument(arena, source),
                },
            });
        },

        else => {
            return error.UnexpectedToken;
        },
    }
}

fn parseTermExpression(self: *Self, arena: *std.heap.ArenaAllocator, source: *tokens.Stream) coral.shaders.ParsingError!*const coral.shaders.Expression {
    const lhs_expression = try self.parseFactorExpression(arena, source);

    if (source.current == .symbol_asterisk) {
        const rhs_expression = try self.parseFactorExpression(arena, source);
        const lhs_type = try lhs_expression.inferType();

        if (lhs_type != try rhs_expression.inferType()) {
            return error.IncompatibleTypes;
        }

        return try create(arena, coral.shaders.Expression{
            .multiply = .{
                .rhs_expression = rhs_expression,
                .lhs_expression = lhs_expression,
                .type = lhs_type,
            },
        });
    }

    if (source.current == .symbol_forward_slash) {
        const rhs_expression = try self.parseFactorExpression(arena, source);
        const lhs_type = try lhs_expression.inferType();

        if (lhs_type != try rhs_expression.inferType()) {
            return error.IncompatibleTypes;
        }

        return try create(arena, coral.shaders.Expression{
            .divide = .{
                .rhs_expression = rhs_expression,
                .lhs_expression = lhs_expression,
                .type = lhs_type,
            },
        });
    }

    return lhs_expression;
}

pub fn parseParameter(self: *Self, arena: *std.heap.ArenaAllocator, source: *tokens.Stream) coral.shaders.ParsingError!?*const coral.shaders.Parameter {
    const parameter = try self.define(arena, coral.shaders.Parameter{
        .identifier = try arena.allocator().dupeZ(u8, try source.current.expectIdentifier()),
        .type = .int,
    });

    parameter.type = try self.parseType(source);

    if (try source.skip(.newline).expectAny(&.{ .symbol_paren_right, .symbol_comma }) == .symbol_comma) {
        parameter.has_next = switch (try source.skip(.newline).expectAny(&.{ .symbol_paren_right, .identifier })) {
            .symbol_paren_right => null,
            .identifier => try self.parseParameter(arena, source),
        };
    }

    return parameter;
}

pub fn parseType(self: *const Self, source: *tokens.Stream) coral.shaders.ParsingError!*const coral.shaders.Type {
    return switch (source.skip(.newline)) {
        .keyword_float => .float,
        .keyword_float4x4 => .float4x4,
        .keyword_float2 => .float2,
        .keyword_float3 => .float3,
        .keyword_float4 => .float4,
        .keyword_int => .int,
        .identifier => |identifier| self.hasGlobal(coral.shaders.Type, identifier) orelse error.UndefinedIdentifier,
        else => error.UnexpectedToken,
    };
}
